import logging
# General
import os
import pickle

import pyresdev.utils as utils
import pyresdev.utils.data_engineering_functions as DEutils

# Data
# ML
# Used in : Data_Engineering

logger = logging.getLogger(__name__)

__all__ = ['preprocessing_training','preprocessing_predictions']




def preprocessing_training_common(df, config, model_type, mapping_dicts=None, imp=None):
    """Common functions used for every campaign in the processing

    Args:
        mapping_dicts: dictionaries for mapping strings to floats
        df ([DataFrame]): [Dataframe to split]
        config ([Dict]): [Config for the campaign]
        model_type(dict): model used ( bv, probabiltiy or multiclass)
    Returns:
        [DataFrames]: [Training and testing divided into dependent and independent variables ]
        [imp]: iterative imputer
    """
    # Data engineering part

    df = (df.pipe(utils.reduce_mem_usage)
          .pipe(DEutils.set_dataframe_columntypes)
          .pipe(DEutils.clean_previous_status, status_to_clean=config['status_to_delete'])
          .pipe(DEutils.clean_rows_with_na, cols_to_clean_na=config['cols_to_clean_na'])
          .pipe(DEutils.clean_companytype)
          .pipe(DEutils.fill_na_with_unknown, columns=config['fill_unknown'])
          .pipe(DEutils.create_company_years_column)
          .pipe(DEutils.fill_seniority_column)
          .pipe(DEutils.filter_countries)
          .pipe(DEutils.clean_companyfollowers_text)
          .pipe(DEutils.convert_kwkey_specialties_and_technologies_to_int)
          .pipe(DEutils.fill_variables)
          .pipe(DEutils.impute_growth)
          .pipe(DEutils.log_transformation, columns=config['cols_to_transform'])
          .pipe(DEutils.drop_columns,columns=config['cols_to_drop'][model_type])
          .pipe(DEutils.correct_C80_success_code)
          )

    #creation and application of the dictionaries
    if mapping_dicts:
        logger.info("No mapping dicts present, creating them")
        mapping_dicts = {}
        mx_prob_dictionary, mx_bv_dictionary = DEutils.create_class_dictionaries(df, "MXType")
        df = DEutils.map_conv_prob_using_dictionary(df, mx_prob_dictionary, mx_bv_dictionary, 'MXType')
        mapping_dicts['mx_prob'] = mx_prob_dictionary
        mapping_dicts['mx_bv'] = mx_bv_dictionary
        state_prob_dictionary, state_bv_dictionary = DEutils.create_class_dictionaries(df, "State")
        df = DEutils.map_conv_prob_using_dictionary(df, state_prob_dictionary, state_bv_dictionary, 'State')
        mapping_dicts['state_prob'] = state_prob_dictionary
        mapping_dicts['state_bv'] = state_bv_dictionary
        industry_prob_dictionary, industry_bv_dictionary = DEutils.create_class_dictionaries(df, "Industry")
        df = DEutils.map_conv_prob_using_dictionary(df, industry_prob_dictionary, industry_bv_dictionary, 'Industry')
        mapping_dicts['industry_prob'] = industry_prob_dictionary
        mapping_dicts['industry_bv'] = industry_bv_dictionary

        mapping_dicts_name = 'mapping_dicts' + config['campaign'] + '.pkl'
        mapping_dicts_output_path = os.path.join('/opt/ml/processing/mapping_dicts', mapping_dicts_name)
        logging.info('Saving dicts to {}'.format(mapping_dicts_output_path))
        pickle.dump(mapping_dicts, open(mapping_dicts_output_path, 'wb'))
    else:
        logging.info("Mapping dicts found! Using them")
        df = DEutils.map_conv_prob_using_dictionary(df,
                                                    mapping_dicts['mx_prob'], mapping_dicts['mx_bv'], 'MXType')

        df = DEutils.map_conv_prob_using_dictionary(df,
                                                    mapping_dicts['state_prob'],mapping_dicts['state_bv'],'State')

        df = DEutils.map_conv_prob_using_dictionary(df,
                                                    mapping_dicts['industry_prob'],mapping_dicts['industry_bv'], 'Industry')

    # Use imputer
    impute_columns = config['impute_columns']
    if imp:
        df = DEutils.impute_nas(df=df, columns_to_impute=impute_columns, imp=imp)
    else:
        df, imp = DEutils.impute_nas_and_save_imputer(df=df, columns_to_impute=impute_columns)
        imputer_name = 'imputer_' + config['campaign'] + '.pkl'
        imputer_output_path = os.path.join('/opt/ml/processing/imputer', imputer_name)
        logging.info('Saving iterative imputer to {}'.format(imputer_output_path))
        pickle.dump(imp, open(imputer_output_path, 'wb'))

    return df, imp, mapping_dicts


def preprocessing_training(df, config, model_type, mapping_dicts=None, imp=None):
    """[Pre-processing function to prepare the data for training the model and to make predictions]
s
    Args:
        model_type:
        config:
        df:
        df ([Pandas Dataframe]): [Pandas Dataframe with the raw data ]
        industry_dictionary ([Dict]): [Dictionary to reduce the cardinality of the "Industry attribute"]
        config(dict) : Dictionary with the configuration ( read from json)
        model_type(string) : type of the model to be used
        mx_bv_dictionary(string): Dictionary of MX converting to bv performance
        mx_prob_dictionary(string): Dictionary of MX converting to prob performance
        model_type(string) : can be probability, bv or multiclass
    Returns:
        df_clean[Pandas Dataframe]: Pre-processed DataFrame ready for training purposes
        :param mx_prob_dictionary:
        :param use_one_hot_encoding:
        :param mapping_dicts:
    """

    # Make a copy of the dataset
    df_leads_clean = df.copy()

    nna = df_leads_clean.isna().sum()
    logging.info("NA in original dataframe:")
    logging.info(f"{nna}")

    # Convert to categories and 
    df_leads_clean, imp = preprocessing_training_common(df=df_leads_clean,
                                                        config=config,
                                                        mapping_dicts=mapping_dicts,
                                                        model_type=model_type,
                                                        imp=imp)

    df_leads_clean = utils.reduce_mem_usage_category(df=df_leads_clean)

    df_leads_clean = DEutils.one_hot_encoding(df=df_leads_clean)

    df_leads_clean = DEutils.convert_date(df=df_leads_clean)

    df_leads_clean = DEutils.set_target_column(df=df_leads_clean, model_type=model_type)

    nna = df_leads_clean.isna().sum()
    logging.info("NA in final dataframe:")
    logging.info(f"{nna}")

    return df_leads_clean, imp


def preprocessing_predictions(df,imputer,mapping_dicts,config):

    df = (df.pipe(DEutils.set_dataframe_columntypes)
          .pipe(DEutils.clean_previous_status, status_to_clean=config['status_to_delete'])
          .pipe(DEutils.clean_rows_with_na, cols_to_clean_na=config['cols_to_clean_na'])
          .pipe(DEutils.clean_companytype)
          .pipe(DEutils.fill_na_with_unknown, columns=config['fill_unknown'])
          .pipe(DEutils.create_company_years_column)
          .pipe(DEutils.fill_seniority_column)
          .pipe(DEutils.filter_countries)
          .pipe(DEutils.clean_companyfollowers_text)
          .pipe(DEutils.convert_kwkey_specialties_and_technologies_to_int)
          .pipe(DEutils.fill_variables)
          .pipe(DEutils.impute_growth)
          .pipe(DEutils.log_transformation, columns=config['cols_to_transform'])
          .pipe(DEutils.drop_columns,columns=config['cols_to_drop'][model_type])
          .pipe(DEutils.correct_C80_success_code)
          )

    return df


import logging
import json
import pickle
import os


logger = logging.getLogger(__name__)

__all__ = ['read_config_file', 'import_dictionary']


def import_dictionary(path, file):
    with open(os.path.join(path, file), 'rb') as f:
        logging.info(f'reading input dict from {path}')
        dict = pickle.load(f)
        return dict


def read_config_file(path):
    try:
        logging.info(f'reading config from {path}')
        with open(path) as file:
            data = file.read()
        data = dict(json.loads(data))
        logging.info(f'accessing parameters')
        config = data['parameters']

    except Exception as exception:
        config = None
        error_message = f"Read config file - {repr(exception)}"
        logging.info(error_message)
    return config






from .aws_tools import *
from .memory_reduction import *
from .general_functions import *

__all__ = (aws_tools.__all__ +
           memory_reduction.__all__+
           general_functions.__all__
           )





import logging
import json
import pickle
import os
import pandas as pd
import pyresdev.utils._regex_utils as reutils
import numpy as np

__all__ = ['create_address_pivot','clean_brackets_from_list','clean_brackets_from_string','extract_info_from_summaries']

def create_address_pivot(df): # Extraer en un dataframe un json dentro de otro json

    aux=pd.json_normalize(df[['linkedin_username','emails']].to_dict(orient='records'),record_path='emails',meta=['linkedin_username'])
    aux['type'].fillna('unknown',inplace=True)
    aux_pivot=aux.pivot_table(index=['linkedin_username'],columns=['type'],values=['address'],aggfunc= lambda x: ', '.join(map(str,x)))
    aux_pivot.columns= ['_'.join(col).strip() for col in aux_pivot.columns.values]
    aux_pivot.reset_index(inplace=True)
    return aux_pivot

def clean_brackets_from_list(df,column):
    df.loc[~df[column].isnull(),column]= df.loc[~df[column].isnull(),column].apply(lambda x: ','.join(map(str, x)))
    return df

def clean_brackets_from_string(df,column):
    df.loc[~df[column].isnull(),column] =  df.loc[~df[column].isnull(),column].apply(lambda x: x.replace('[','').replace(']',''))
    return df

def extract_info_from_summaries(df,column,is_list):
    #Convert experience list of dicts into a single string
    column_str=column+'_str'

    if is_list:
        print('column is list! converting to string')
        df[column_str]= df[column].apply(lambda x: ','.join(map(str,x)))
        df[column_str]=df[column_str].astype(str)
    else:
        print('column is not list! converting to string')
        df[column_str]=df[column].astype(str)
    
    # Phone numbers
    out_phone_column='out_phone_in_'+column
    mobile_phone_column='mobile_phone_in_'+column
    office_phone_column='office_phone_in_'+column
    df[out_phone_column],df[mobile_phone_column],df[office_phone_column] = zip(*df[column_str].apply(lambda x: reutils.search_for_phones(str(x))))
    print(f'{column} with phone numbers:')
    print(sum(df[out_phone_column].apply(lambda d: len(d))>0))
    for row in df[out_phone_column]:
        if '1-10' in row:
            row.remove('1-10')
        if '11-50' in row:
            row.remove('11-50')
        if '201-500' in row:
            row.remove('201-500')
        if '1001-5000' in row:
            row.remove('1001-5000')
        if '5001-10000' in row:
            row.remove('5001-10000')
        if '501-1000' in row:
            row.remove('501-1000')
        if '51-200' in row:
            row.remove('51-200')
    print(f'{column} with phone numbers after removing sizes:')
    print(sum(df[out_phone_column].apply(lambda d: len(d))>0))
    df[out_phone_column] = df[out_phone_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[out_phone_column].map(lambda d: len(d)) == 0,out_phone_column]=None
    df.loc[~df[out_phone_column].isnull(),out_phone_column]= df.loc[~df[out_phone_column].isnull(),out_phone_column].apply(lambda x: ','.join(map(str, x)))
    print(f'{column} with mobile phone numbers:')
    print(sum(df[mobile_phone_column].apply(lambda d: len(d))>0))
    df[mobile_phone_column] = df[mobile_phone_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[mobile_phone_column].map(lambda d: len(d)) == 0,mobile_phone_column]=None
    df.loc[~df[mobile_phone_column].isnull(),mobile_phone_column]= df.loc[~df[mobile_phone_column].isnull(),mobile_phone_column].apply(lambda x: ','.join(map(str, x)))
    print(f'{column} with office phone numbers:')
    print(sum(df[office_phone_column].apply(lambda d: len(d))>0))
    df[office_phone_column] = df[office_phone_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[office_phone_column].map(lambda d: len(d)) == 0,office_phone_column]=None
    df.loc[~df[mobile_phone_column].isnull(),mobile_phone_column]= df.loc[~df[mobile_phone_column].isnull(),mobile_phone_column].apply(lambda x: ','.join(map(str, x)))
    
    
    # emails
    mails_column='emails_found_in_'+column
    df[mails_column]= df[column_str].apply(lambda x: reutils.search_for_emails(str(x)))
    print(f'{column} with emails:')
    print(sum(df[mails_column].apply(lambda d: len(d))>0))
    df[mails_column] = df[mails_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[mails_column].map(lambda d: len(d)) == 0,mails_column]=None
    df.loc[~df[mails_column].isnull(),mails_column]= df.loc[~df[mails_column].isnull(),mails_column].apply(lambda x: ','.join(map(str, x)))
    
    # search_for_websites
    websites_column='websites_found_in_'+column
    df[websites_column]= df[column_str].apply(lambda x: reutils.search_for_websites(str(x)))
    print(f'{column} with websites:')
    df[websites_column] = df[websites_column].map(lambda l: list(map(str.strip, l)))
    print(sum(df[websites_column].apply(lambda d: len(d))>0))
    df.loc[df[websites_column].map(lambda d: len(d)) == 0,websites_column]=None
    df.loc[~df[websites_column].isnull(),websites_column]= df.loc[~df[websites_column].isnull(),websites_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_twitter_usernames
    twitter_column='twitter_found_in_'+column
    df[twitter_column]= df[column_str].apply(lambda x: reutils.search_for_twitter_usernames(str(x)))
    print(f'{column} with twitter:')
    print(sum(df[twitter_column].apply(lambda d: len(d))>0))
    df[twitter_column] = df[twitter_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[twitter_column].map(lambda d: len(d)) == 0,twitter_column]=None
    df.loc[~df[twitter_column].isnull(),twitter_column]= df.loc[~df[twitter_column].isnull(),twitter_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_facebook_profiles
    facebook_column='facebook_found_in_'+column
    df[facebook_column]= df[column_str].apply(lambda x: reutils.search_for_facebook_profiles(str(x)))
    print(f'{column} with facebook:')
    print(sum(df[facebook_column].apply(lambda d: len(d))>0))
    df[facebook_column] = df[facebook_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[facebook_column].map(lambda d: len(d)) == 0,facebook_column]=None
    df.loc[~df[facebook_column].isnull(),facebook_column]= df.loc[~df[facebook_column].isnull(),facebook_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_github_usernames
    github_column='github_found_in_'+column
    df[github_column]= df[column_str].apply(lambda x: reutils.search_for_github_usernames(str(x)))
    print(f'{column} with github:')
    print(sum(df[github_column].apply(lambda d: len(d))>0))
    df[github_column] = df[github_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[github_column].map(lambda d: len(d)) == 0,github_column]=None
    df.loc[~df[github_column].isnull(),github_column]= df.loc[~df[github_column].isnull(),github_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_instagram_profiles
    instagram_column='instagram_found_in_'+column
    df[instagram_column]=np.nan
    # df[instagram_column]= df[column_str].apply(lambda x: reutils.search_for_instagram_profiles(str(x)))
    # print(f'{column} with iinstagram:')
    # print(sum(df[instagram_column].apply(lambda d: len(d))>0))
    # df[instagram_column] = df[instagram_column].map(lambda l: list(map(str.strip, l)))
    # df.loc[df[instagram_column].map(lambda d: len(d)) == 0,instagram_column]=None
    # df.loc[~df[instagram_column].isnull(),instagram_column]= df.loc[~df[instagram_column].isnull(),instagram_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_linkedin_profiles
    linkedin_column='linkedin_found_in_'+column
    df[linkedin_column]=np.nan
    # df[linkedin_column]= df[column_str].apply(lambda x: reutils.search_for_linkedin_profiles(str(x)))
    # print(f'{column} with linkedin:')
    # print(sum(df[linkedin_column].apply(lambda d: len(d))>0))
    # df[linkedin_column] = df[linkedin_column].map(lambda l: list(map(str.strip, l)))
    # df.loc[df[linkedin_column].map(lambda d: len(d)) == 0,linkedin_column]=None
    # df.loc[~df[linkedin_column].isnull(),linkedin_column]= df.loc[~df[linkedin_column].isnull(),linkedin_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_linkedin_company
    linkedin_company_column='linkedin_company_found_in_'+column
    df[linkedin_company_column]= df[column_str].apply(lambda x: reutils.search_for_linkedin_company(str(x)))
    print(f'{column} with linkedin company:')
    print(sum(df[linkedin_company_column].apply(lambda d: len(d))>0))
    df[linkedin_company_column] = df[linkedin_company_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[linkedin_company_column].map(lambda d: len(d)) == 0,linkedin_company_column]=None
    df.loc[~df[linkedin_company_column].isnull(),linkedin_company_column]= df.loc[~df[linkedin_company_column].isnull(),linkedin_company_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_medium_user
    medium_column='medium_found_in_'+column
    df[medium_column] = np.nan
    # df[medium_column]= df[column_str].apply(lambda x: reutils.search_for_medium_user(str(x)))
    # print(f'{column} with medium:')
    # print(sum(df[medium_column].apply(lambda d: len(d))>0))
    # df[medium_column] = df[medium_column].map(lambda l: list(map(str.strip, l)))
    # df.loc[df[medium_column].map(lambda d: len(d)) == 0,medium_column]=None
    # df.loc[~df[medium_column].isnull(),medium_column]= df.loc[~df[medium_column].isnull(),medium_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_skype
    skype_column='skype_found_in_'+column
    df[skype_column]= df[column_str].apply(lambda x: reutils.search_for_skype(str(x)))
    print(f'{column} with skype:')
    print(sum(df[skype_column].apply(lambda d: len(d))>0))
    df[skype_column] = df[skype_column].map(lambda l: list(map(str.strip, l)))
    df.loc[df[skype_column].map(lambda d: len(d)) == 0,skype_column]=None
    df.loc[~df[skype_column].isnull(),skype_column]= df.loc[~df[skype_column].isnull(),skype_column].apply(lambda x: ','.join(map(str, x)))
    
    # reutils.search_for_telegram
    telegram_column='telegram_found_in_'+column
    df[telegram_column]=np.nan
    # df[telegram_column]= df[column_str].apply(lambda x: reutils.search_for_telegram(str(x)))
    # print(f'{column} with telegram:')
    # print(sum(df[telegram_column].apply(lambda d: len(d))>0))
    # df[telegram_column] = df[telegram_column].map(lambda l: list(map(str.strip, l)))
    # df.loc[df[telegram_column].map(lambda d: len(d)) == 0,telegram_column]=None

    return df





# General
import datetime
import os
import pickle
import warnings
import argparse
import logging
import json
from datetime import date
from math import log
# Data
import pandas as pd
import numpy as np
import regex as re
# ML
# Used in : Data_Engineering
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer  # Used in : Data_Engineering
from sklearn.linear_model import BayesianRidge  # Used in : Data_Engineering
from sklearn.exceptions import DataConversionWarning

logger = logging.getLogger(__name__)


def set_dataframe_columntypes(df):
    """Manual setting of the datatypes

    Args:
        df ([dataframe]): [dataframe with the leads information]

    Returns:
       df [df]: [Dataframe with the leads information, with the new types]
    """
    logger.info("Changing data types")
    df = clean_rows_with_na(df=df, cols_to_clean_na=['IDEmailStatus'])
    if 'idLead' in df:
        df['idLead'] = df['idLead'].astype('int32')
    if 'idCampaign' in df:
        df['idCampaign'] = df['idCampaign'].astype('Int64')
        df['idCampaign'] = df['idCampaign'].apply(str)
    if 'BulkStatus' in df:
        df['BulkStatus'] = df['BulkStatus'].astype('Int64')
        df['BulkStatus'] = df['BulkStatus'].apply(str)
    if 'campaign_bulk_date' in df:
        df['campaign_bulk_date'] = df['campaign_bulk_date'].astype('object')
        df['campaign_bulk_date'] = pd.to_datetime(df['campaign_bulk_date'])
        df['campaign_bulk_date'] = df['campaign_bulk_date'].dt.date
    if 'IDEmailStatus' in df:
        df['IDEmailStatus'] = df['IDEmailStatus'].astype('int8')
    if 'Ranking' in df:
        df['Ranking'] = df['Ranking'].astype('float32')
    if 'BV' in df:
        df['BV'] = df['BV'].astype('float32')
    if 'Size' in df:
        df['Size'] = df['Size'].astype('object')
    if 'Industry' in df:
        df['Industry'] = df['Industry'].astype('object')
    if 'CompanyFollowers' in df:
        df['CompanyFollowers'] = df['CompanyFollowers'].astype('object')
    if 'QEmployeesOnLinkedIn' in df:
        df['QEmployeesOnLinkedIn'] = df['QEmployeesOnLinkedIn'].astype('float64')
    if 'Type' in df:
        df['Type'] = df['Type'].astype('object')
    if 'Specialties' in df:
        df['Specialties'] = df['Specialties'].astype('object')
    if 'MXType' in df:
        df['MXType'] = df['MXType'].astype('object')
    if 'Technologies' in df:
        df['Technologies'] = df['Technologies'].astype('object')
    if 'Country' in df:
        df['Country'] = df['Country'].astype('object')
    if 'State' in df:
        df['State'] = df['State'].astype('object')
    if 'Title' in df:
        df['Title'] = df['Title'].astype('object')
    if 'kw_key' in df:
        df['kw_key'] = df['kw_key'].astype('object')
    if 'Seniority' in df:
        df['Seniority'] = df['Seniority'].astype('object')
    if 'Department' in df:
        df['Department'] = df['Department'].astype('object')
    if 'LeadConnections' in df:
        df['LeadConnections'] = df['LeadConnections'].astype('float64')
    if 'LeadFollowers' in df:
        df['LeadFollowers'] = df['LeadFollowers'].astype('float64')
    if 'PreviousContacts' in df:
        df['PreviousContacts'] = df['PreviousContacts'].astype('float16')
    if 'PreviousStatus' in df:
        df['PreviousStatus'] = df['PreviousStatus'].astype('float16')
    if 'DaysSinceLastContact' in df:
        df['DaysSinceLastContact'] = df['DaysSinceLastContact'].astype(
            'float16')
    if 'DaysSinceFirstContact' in df:
        df['DaysSinceFirstContact'] = df['DaysSinceFirstContact'].astype(
            'float16')
    if 'DaysSincePositionStart' in df:
        df['DaysSincePositionStart'] = df['DaysSincePositionStart'].astype(
            'float16')

    logger.info("Datatypes after modification")
    logger.info(f'{df.dtypes}')


    return df

def one_hot_encoding(df):
    """Generate one hot encoded dataset
    WARNING:
        Only encodes colums with category

    Args:
        df ([dataframe]): [dataframe]

    Returns:
        df[dataframe]: [dataframe with categorical variables one hot encoded]
    """
    # These variables must not be categories
    if 'campaign_bulk_date' in df.columns:
        df['campaign_bulk_date'] = df['campaign_bulk_date'].astype('object')
    if 'IDEmailStatus' in df.columns:
        df['IDEmailStatus'] = df['IDEmailStatus'].astype('object')
    logging.info("Data Types before encoding:")
    logging.info(f"{df.dtypes}")
    df = pd.get_dummies(data=df,
                        columns=df.select_dtypes(include=['category']).columns.tolist())
    logging.info("Data Types after encoding:")
    logging.info(f"{df.dtypes}")
    return df


def convert_date(df):
    # Convert date
    if 'campaign_bulk_date' in df.columns:
        df['campaign_bulk_date'] = df['campaign_bulk_date'].astype('object')
        df['campaign_bulk_date'] = pd.to_datetime(df['campaign_bulk_date'])
        df['campaign_bulk_date'] = df['campaign_bulk_date'].dt.date
    else:
        raise Exception('campaign_bulk_date not present! unable to perform dataset split')
    return df

def drop_columns(df, columns):
    """Drops the selected columns

    Args:
        df ([dataframe]): [leads dataframe]
        columns ([list]): [list columns to drop]

    Returns:
        [df]: [leads dataframe]
    """
    for column in columns:
        if column in df:
            logging.info(f'dropping column : {column}')
            df = df.drop(column, axis=1)
    return df


def clean_previous_status(df, status_to_clean):
    """Eliminate rows with a defined previous status value
    Args:
        df ([dataframe]): [dataframe with the leads]
        status_to_clean ([list of ints]): [list with the status to clean]

    Returns:
        df [dataframe]: [dataframe with the leads]
    """

    logger.info(f"Cleaning invalid previous status ( {status_to_clean})")
    for status in status_to_clean:
        if 'PreviousStatus' in df:
            df = df[df.PreviousStatus != status]
    return df


def clean_rows_with_na(df, cols_to_clean_na):
    """Eliminates rows with NAs in Attributes with low % NA Percentage

    Args:
        df ([dataframe]): Leads dataframe
    Returns:
        df [dataframe]: [dataframe with the leads]
    """
    # logger  amount of rows
    nrows_pre = df.shape[0]
    logger.info(f"Original rows: {nrows_pre}")

    for col in cols_to_clean_na:
        if col in df:
            nas_sum = df[col].isna().sum()
            logger.info(f"NAs in {col}: {nas_sum}")
            df = df[df[col].notna()]
    nrows_post = df.shape[0]
    percentage_eliminated = (nrows_pre - nrows_post) / (nrows_pre * 100 + 1)
    logger.info(f" Rows after deleting rows with NA: {nrows_post}, Percentage: {percentage_eliminated}")

    return df


def fill_na_with_unknown(df, columns):
    """[summary]
    Args:
        df ([dataframe]): [dataframe with the leads]
        columns ([list of strings]): [list with the names of the columns]
    Returns:
        df [dataframe]: [dataframe with the leads]
    """
    for col in columns:
        if col in df.columns:
            logger.info(f'filling NA in {col} with "unknown" ')
            df[col] = df[col].fillna('Unknown')
    return df


def correct_C80_success_code(df):
    # C80 has an special conversion Status, setting it to 14 like the other campaigns

    # Correcting non perfect conversions
    df.loc[(df['idCampaign'] == '80') &
           (df['BV'].notna()), 'IDEmailStatus'] == 2
    aux = df.loc[(df['idCampaign'] == '80') &
                 (df['IDEmailStatus'] == 31)].shape
    logger.info(f'C80 successes: {aux}')
    df.loc[(df['idCampaign'] == '80') &
           (df['IDEmailStatus'] == 31),
           'IDEmailStatus'] = 14
    aux = df.loc[(df['idCampaign'] == '80') &
                 (df['IDEmailStatus'] == 14)].shape
    logger.info(f'C80 successes: {aux}')

    return df





def set_idemailstatus_target_column(df):
    if ('IDEmailStatus' in df.columns):
        logger.info('Setting E-Mail status target ')
        df = df[~df['IDEmailStatus'].isin([3, 5, 7, 32])]
        df.loc[df['IDEmailStatus'] != 14, 'IDEmailStatus'] = 0
        df.loc[df['IDEmailStatus'] == 14, 'IDEmailStatus'] = 1
        date_col = df['campaign_bulk_date']
        df.drop(labels=['campaign_bulk_date'], axis=1, inplace=True)
        df.insert(loc=0, value=date_col, column='campaign_bulk_date')
        status_col = df['IDEmailStatus']
        df.drop(labels=['IDEmailStatus'], axis=1, inplace=True)
        df.insert(loc=1, value=status_col, column='IDEmailStatus')
        bv_col = df['BV']
        df.drop(labels=['BV'], axis=1, inplace=True)
        df.insert(loc=2, value=bv_col, column='BV')

    else:
        logger.error('No IDEmailSTatus column found')
        raise Exception('No IDEmailStatus column found!!!!!')
    return df


def set_bv_target_column(df):
    if ('BV' in df.columns):
        logger.info('Setting BV status target ')
        df = df[df.BV.notna()]
        df['BV'] = df['BV'].astype('float32')
        date_col = df['campaign_bulk_date']
        df.drop(labels=['campaign_bulk_date'], axis=1, inplace=True)
        df.insert(loc=0, value=date_col, column='campaign_bulk_date')
        bv_col = df['BV']
        df.drop(labels=['BV'], axis=1, inplace=True)
        df.insert(loc=1, value=bv_col, column='BV')

    else:
        logger.error('No BV column found')
        raise Exception('No BV column found!!!!!')
    return df


def set_bv_fake_column(df):
    if ('BV' in df.columns):
        logger.info('Setting BV fake target ')
        n = len(df)
        df['BV'] = np.ones(n)
        date_col = df['campaign_bulk_date']
        df.drop(labels=['campaign_bulk_date'], axis=1, inplace=True)
        df.insert(loc=0, value=date_col, column='campaign_bulk_date')
        bv_col = df['BV']
        df.drop(labels=['BV'], axis=1, inplace=True)
        df.insert(loc=1, value=bv_col, column='BV')

    else:
        logger.error('No BV column found')
        raise Exception('No BV column found!!!!!')
    return df


def set_bv_discrete_target_column(df, bins):
    if ('BV' in df.columns):
        logger.info('Setting BV status target for classification ')
        df.BV.fillna(value=0, inplace=True)
        labels = [0, 1, 2, 3, 4, 5]
        df.insert(loc=1, column='BV_cut', value=pd.cut(x=df['BV'], bins=bins, labels=labels).astype('float32'))
    else:
        logger.error('No BV column found')
        raise Exception('No BV column found!!!!!')
    return df


def set_target_column(df, model_type):
    if model_type == 'probability':
        df = set_idemailstatus_target_column(df=df)
    elif model_type == 'bv':
        df = set_bv_target_column(df=df)
    elif model_type == 'bvfake':
        df = set_bv_fake_column(df=df)
    else:
        raise Exception('Incorrect model_type !!!!!')
    return df


def create_company_years_column(df):
    """Year founded, creation of the Company years category using Year founded

    Args:
        df ([dataframe]): [dataframe with the leads]

    Returns:
        df [dataframe]: [dataframe with the leads]
    """

    logger.info("Creating company years category")
    # Create Company years category
    if df['YearFounded'].dtype == object:
        df['YearFounded'] = df['YearFounded'].str.extract(
            '(\d{4})$', expand=False).astype('Float16')

    CurrentYear = date.today().year
    df['YearFounded'] = df['YearFounded'].fillna(0)
    df['CompanyYears'] = CurrentYear - df['YearFounded']
    df.loc[(df['CompanyYears'] < 0) | (df['CompanyYears'] > 250), 'CompanyYears'] = np.nan
    df = df.drop(['YearFounded'], axis=1)
    return df


def clean_companytype(df):
    df.loc[
        (df['Type'] != 'Educational Institution') &
        (df['Type'] != 'Government Agency') &
        (df['Type'] != 'Nonprofit') &
        (df['Type'] != 'Partnership') &
        (df['Type'] != 'Privately Held') &
        (df['Type'] != 'Public Company') &
        (df['Type'] != 'Self-Employed'),
        'Type'] = 'Unknown'
    return df


def fill_seniority_column(df):
    """Defines new cases for Seniority and sets the rest to 'other'

    Args:
        df ([dataframe]): [Dataframe with the leads]
    Returns:
        df [datafrane]: [dataframe with the leads]
    """
    logger.info("Filling Seniority from some Titles")
    df.loc[(df['Title'].str.contains('Specialist', 'Title')) &
           (df['Seniority'].isna()), 'Seniority'] = 'Specialist'
    df.loc[(df['Title'].str.contains('Associate', 'Title')) &
           (df['Seniority'].isna()), 'Seniority'] = 'Associate'
    df['Seniority'] = df['Seniority'].fillna('other')

    return df


def filter_countries(df):
    logger.info(f'countries out of top 8 go to other')
    top = df['Country'].value_counts().head(8).index.tolist()
    df.loc[~df['Country'].isin(top), 'Country'] = 'other'
    return df


def process_bulkstatus(df):
    logger.info("Processing Bulk Status column")
    if 'BulkStatus' in df:
        logger.info(f"{df['BulkStatus'].value_counts()}")
        df['is_OOO'] = 0
        df.loc[(df['BulkStatus'] == 6), 'is_OOO'] = 1
        df['is_NN'] = 0
        df.loc[(df['BulkStatus'] == 12), 'is_NN'] = 1
        df.drop(labels=['BulkStatus'], axis=1, inplace=True)
        logger.info('Deleted BulkStatus column')
        logger.info('OOO')
        logger.info(df['is_OOO'].value_counts())
        logger.info('NN')
        logger.info(df['is_NN'].value_counts())
    else:
        logger.info('Column not present')
    return df


def impute_growth(df):
    logger.info("Applying growth imputation")
    df['Growth12M'] = df['Growth12M'].fillna(df['Growth6M']*2)
    df['Growth6M'] = df['Growth6M'].fillna(df['Growth3M']*2)
    df['Growth3M'] = df['Growth3M'].fillna(df['Growth6M']/2)

    return df


def convert_kwkey_specialties_and_technologies_to_int(df):
    """Using regular expressions, converts a comma separated list of keys and technologies into the number of elements of this columns

    Args:
        df ([dataframe]): [Dataframe with the leads]

    Returns:
       df [dataframe]: [Dataframe with the leads]
    """
    logger.info('Counting KeyWords')
    # number of elements in kw_key
    df['Nkw_key'] = df.kw_key.apply(
        lambda x: 0 if pd.isnull(x) else len(re.findall("(?:\".*?\"|\S)+", x)))
    # Number of elements in technologies
    logger.info('Counting Technologies')
    df['NTechnologies'] = df.Technologies.apply(
        lambda x: 0 if pd.isnull(x) else len(re.split(',', x)))
    logger.info('Counting Specialties')
    df['NSpecialties'] = df.Technologies.apply(
        lambda x: 0 if pd.isnull(x) else len(re.split(',', x)))

    return df


def fill_variables(df):
    """fill variables with Nulls
    Args:
        df ([dataframe]): [Dataframe with the leads]

    Returns:
        df [dataframe]: [Dataframe with the leads]
    """

    variables_dict = {
        'PreviousContacts': 0,
        'DaysSinceLastContact': 0,
        'DaysSinceFirstContact': 0,
        'NumFoundingRounds': 0,
        'FundingAmount': 0,
        'DaysSinceLastSeed': 99999
    }

    for key in variables_dict.keys():
        if key in df:
            logger.info(f'Filling variable {key} nulls with {variables_dict[key]}')
            logger.info(df[key].isna().sum())
            df[key] = df[key].fillna(variables_dict[key])
        else:
            logger.info(f'{key} not available to fill')
    return df


def impute_nas_and_save_imputer(df, columns_to_impute):
    df_impute = df[columns_to_impute]
    logger.info(f"Imputing NAs in {columns_to_impute}")
    logger.info("NAs in DataFrame before imputation:")
    nna = df_impute.isna().sum()
    logger.info(f"{nna}")

    logger.info("No imputer provided, instancing and fitting a new one")
    imp = IterativeImputer(missing_values=np.nan,
                           sample_posterior=False,
                           max_iter=10,
                           tol=0.01,
                           initial_strategy='mean',
                           estimator=BayesianRidge())

    imp.fit(df_impute)
    logger.info("Creating new imputer")
    df_impute = pd.DataFrame(data=imp.transform(df_impute),
                             columns=df_impute.columns)

    logger.info("NAs in DataFrame after imputation:")
    nna = df_impute.isna().sum()
    logger.info(f"{nna}")

    # Assign column
    for column in columns_to_impute:
        df[column] = df_impute[column].to_numpy()
    del df_impute

    return df, imp


def impute_nas(df, columns_to_impute, imp):
    """Imputes NAs in fields of interest, using sklearn's Iterative imputer

    Args:
        df ([dataframe]): [Dataframe with the leads]
        columns_to_impute ([list of columns]): [List of strings with the columns to impute]

    Returns:
        df [dataframe]: [Dataframe with the leads]
    """
    df_impute = df[columns_to_impute]
    logger.info(f"Imputing NAs in {columns_to_impute}")
    logger.info("NAs in DataFrame before imputation:")
    nna = df_impute.isna().sum()
    logger.info(f"{nna}")
    # Instance the imputer if it is not created

    print("Using provided imputer")
    df_impute = pd.DataFrame(data=imp.transform(df_impute),
                             columns=df_impute.columns)

    logger.info("NAs in DataFrame after imputation:")
    nna = df_impute.isna().sum()
    logger.info(f"{nna}")

    # Assign column
    for column in columns_to_impute:
        df[column] = df_impute[column].to_numpy()
    del df_impute

    return df


def log_transformation(df, columns):
    """Transform selected variables using log, imputing mean in negative values to avoid errors

    Args:
        df ([dataframe]): [datafram with the leads]
        columns ([list of strings]): [list of the columns to transform]
    Returns:
        df ([dataframe]): [datafram with the leads]
    """
    for column in columns:
        logger.info(f'Log Transforming : {column}')
        df.loc[(df[column] < 0), column] = df[column].mean()
        df[column] = df[column].apply(lambda x: np.log10(x + 1))
    return df


def clean_companyfollowers_text(df):
    """Deletes any word in the CompanyFollowers field, and converts to float

    Args:
        df ([dataframe]): dataframe with the leads

    Returns:
        df[dataframe]: [dataframe with the leads]
    """

    df['CompanyFollowers'] = df['CompanyFollowers'].str.extract(
        '(\d+)', expand=False).astype('float')
    return df


def clean_mxtype(df, mxtypes):
    """

    :param df: Input Dataset
    :param mxtypes: MXTypes to maintain ( the rest will go to others)
    :return df: processed Dataset
    """

    if 'MXType' not in df.columns:
        raise Exception("MXType column not preset ( Check Caps)")
    logger.info('Defining MXTypes')
    logger.info(f'legal MXTypes: {mxtypes}')
    df.loc[~df['MXType'].isin(mxtypes), 'MXType'] = 'Others'
    logger.info(f'{df.MXType.value_counts()}')
    return df


def create_class_dictionaries(df_original,class_to_use:str):
    logger.info(f'Creating dictionaries for {class_to_use}')
    df=df_original[[class_to_use, 'BV', 'IDEmailStatus']].copy()
    df['BV']=df['BV'].fillna(0)
    df.loc[(df['IDEmailStatus'] != 14) &
             (df['IDEmailStatus'] != 31), 'IDEmailStatus'] = 0
    df.loc[df['IDEmailStatus'] == 14, 'IDEmailStatus'] = 1
    df.loc[df['IDEmailStatus'] == 31, 'IDEmailStatus'] = 1

    logger.info(f"{df['BV'].value_counts(dropna=False)}")
    groupby = df.groupby(class_to_use).agg(BV=('BV', 'mean'),
                                           Conversions=('IDEmailStatus', 'mean')
                                           )
    groupby = groupby.reset_index()
    logger.info(f"{groupby.head()}")
    list_dict = groupby.to_dict(orient='records')
    conversion_dict = {}
    bv_dict = {}
    for element in list_dict:
        conversion_dict[f'{element[class_to_use]}'] = element['Conversions']
        bv_dict[f'{element[class_to_use]}'] = element['BV']
    return conversion_dict,bv_dict

def clean_mxtype_using_dictionary(df,mx_bv_dict,mx_prob_dict):
    """

    :param df: Input Dataset
    :param mx_bv_dict: MXTypes dictionary by bv performance
    :param mx_prob_dict: MXTypes dictionary by prob performance
    :return df: processed Dataset
    """

    if 'MXType' not in df.columns:
        raise Exception ("MXType column not preset ( Check Caps)")
    logger.info('Defining MXTypes')
    df['mx_bv'] = df['MXType']
    df = df.replace({'mx_bv': mx_bv_dict})
    logger.info(f'{df.mx_bv.value_counts()}')
    df['mx_prob'] = df['MXType']
    df = df.replace( {'mx_prob': mx_prob_dict} )
    logger.info(f'{df.mx_prob.value_counts()}')
    return df


def map_conv_prob_using_dictionary(df, bv_dict, prob_dict, column:str):
    """

    :param df: Input Dataset
    :param mx_bv_dict: MXTypes dictionary by bv performance
    :param mx_prob_dict: MXTypes dictionary by prob performance
    :return df: processed Dataset
    """

    if column not in df.columns:
        raise Exception("MXType column not preset ( Check Caps)")
    logger.info(f'mapping bv and prob dicts for {column}')
    bv_col = column+"_bv"
    df[bv_col] = df[column].map(bv_dict)
    prob_col = column+"_prob"
    df[prob_col] = df[column].map(prob_dict)
    df.drop(column,axis=1,inplace=True)
    return df


def create_new_position_flag(df, days):
    if 'DaysSincePositionStart' in df:
        logging.info(f'Days Since position detected, creating flag for {days} days')
        df['DaysSincePositionStart'].fillna(999)
        colname = 'new_position_' + str(days)
        df[colname] = 0
        df.loc[df['DaysSincePositionStart'] < days, colname] = 1
        logging.info(f'{df[colname].value_counts()}')
    else:
        logging.info(' dayssincepositionstart Column not found, no changes were made.')
    return df



import logging
import json
import pickle
import os
import pandas as pd
import datetime
logger = logging.getLogger(__name__)

"""

Old functions that should not be again used are saved here and excluded from the init file. They remain only for logging
purposes

"""

def correct_C90_dates(df):
    # C90 changed it's configuration in may 2021, so we need to eliminate records before
    C90_start_date = datetime.datetime.strptime('2021-05-01', '%Y-%m-%d').date()
    logger.info(f'shape before C90 date correction: {df.shape}')
    df = df.loc[~ ((df['idCampaign'] == '90') & (df['campaign_bulk_date'] < C90_start_date))]
    logger.info(f'shape after C90 date correction: {df.shape}')
    return df

def create_month_status(df):
    logger.info("Creating Month of the date column")
    if 'campaign_bulk_date' in df.columns:
        df['MonthStatus'] = pd.DatetimeIndex(
            df['campaign_bulk_date']).month.astype('int16')
        df['MonthStatus'] = df['MonthStatus'].astype("category")
    return df